import pandas as pd
import numpy as np
import networkx as nx

df_columns = ["id", "f_degree_in", "f_degree_out", "f_amount_in", "f_amount_out",
              "f_nr_trans_in", "f_nr_trans_out", "f_ego_nr_nodes", "f_ego_nr_edges",
              "f_egored_degree_in", "f_egored_degree_out",
              "f_egored_amount_in", "f_egored_amount_out",
              "f_egored_nr_trans_in", "f_egored_nr_trans_out",
              "f_egored_nr_nodes", "f_egored_nr_edges", "nr_alerts", "nr_reports"]

df_features = ["id", "f_degree_in", "f_degree_out", "f_amount_in", "f_amount_out",
               "f_average_amount_in", "f_average_amount_out", "f_ego_edge_density",
               "f_egored_degree_in_rel", "f_egored_degree_out_rel",
               "f_egored_amount_in_rel", "f_egored_amount_out_rel",
               "f_egored_average_amount_in", "f_egored_average_amount_out", "f_egored_edge_density"]


def exist_common_neighbors(n: int, a_list: list, G: nx.DiGraph):
    """Check if the neighbors of n are in a_list

    Parameters
    ----------
    n : integer
        node in G
        
    a_list : list of integers
        neighbors of another node, with which n is connected
        
    G : Networkx graph
        the graph
        
    Returns
    -------
    bool
        True if there is at least a common neighbor
    """
    my_neighbors = G.successors(n)
    if set(my_neighbors) & set(a_list):
        return True
    else:
        my_neighbors = G.predecessors(n)
        if set(my_neighbors) & set(a_list):
            return True
        else:
            return False
    

def features_from_basic_14(X_basic: pd.DataFrame, X: pd.DataFrame, ii):
    """Compute the 14 egonet and egored features from the basic features,
       for a single row (the same!) of the dataframes
       
    Parameters
    ----------
    X_basic : pd.DataFrame
        Dataframe with basic features.
    X : pd.DataFrame
        Dataframe with computed features.
    ii : integer
        Row index where the modification is made.

    Returns
    -------
    None, but X is modified.

    """
    # copy parameters that are the same and compute the others
    # egonet
    X["f_degree_in"].iat[ii] = X_basic["f_degree_in"].iat[ii]
    X["f_degree_out"].iat[ii] = X_basic["f_degree_out"].iat[ii]
    X["f_amount_in"].iat[ii] = X_basic["f_amount_in"].loc[ii]
    X["f_amount_out"].iat[ii] = X_basic["f_amount_out"].iat[ii]
    if int(X_basic["f_nr_trans_in"].iat[ii]) > 0:
        X["f_average_amount_in"].iat[ii] = X_basic["f_amount_in"].iat[ii] / \
                X_basic["f_nr_trans_in"].iat[ii]
    if int(X_basic["f_nr_trans_out"].iat[ii]) > 0:
        X["f_average_amount_out"].iat[ii] = X_basic["f_amount_out"].iat[ii] / \
                X_basic["f_nr_trans_out"].iat[ii]
    X["f_ego_edge_density"].iat[ii] = X_basic["f_ego_nr_edges"].iat[ii] / X_basic["f_ego_nr_nodes"].iat[ii]

    # egored
    if int(X_basic["f_degree_in"].iat[ii]) > 0:
        X["f_egored_degree_in_rel"].iat[ii] = X_basic["f_egored_degree_in"].iat[ii] / \
                X_basic["f_degree_in"].iat[ii]
    if int(X_basic["f_degree_out"].iat[ii]) > 0:
        X["f_egored_degree_out_rel"].iat[ii] = X_basic["f_egored_degree_out"].iat[ii] / \
                X_basic["f_degree_out"].iat[ii]
    if float(X_basic["f_amount_in"].iat[ii]) > 0:
        X["f_egored_amount_in_rel"].iat[ii] = X_basic["f_egored_amount_in"].iat[ii] / \
                X_basic["f_amount_in"].iat[ii]
    if float(X_basic["f_amount_out"].iat[ii]) > 0:
        X["f_egored_amount_out_rel"].iat[ii] = X_basic["f_egored_amount_out"].iat[ii] / \
                X_basic["f_amount_out"].iat[ii]
    if int(X_basic["f_egored_nr_trans_in"].iat[ii]) > 0:
        X["f_egored_average_amount_in"].iat[ii] = X_basic["f_egored_amount_in"].iat[ii] / \
                X_basic["f_egored_nr_trans_in"].iat[ii]
    if int(X_basic["f_egored_nr_trans_out"].iat[ii]) > 0:
        X["f_egored_average_amount_out"].iat[ii] = X_basic["f_egored_amount_out"].iat[ii] / \
                X_basic["f_egored_nr_trans_out"].iat[ii]
    if int(X_basic["f_egored_nr_nodes"].iat[ii]) > 0:
        X["f_egored_edge_density"].iat[ii] = X_basic["f_egored_nr_edges"].iat[ii] / \
                X_basic["f_egored_nr_nodes"].iat[ii]


def online_update_features_and_graph(G: nx.DiGraph, t_df: pd.DataFrame, f_df: pd.DataFrame, X: pd.DataFrame):
    """Update basic features, egonet & egored features (for anomaly detection) and graph
    
    Parameters
    ----------
    G : networkx DiGraph
        Graph of all past transactions.
    t_df : pd.DataFrame
        Transaction to be processed.
    f_df : pd.DataFrame
        Basic features (corresponding to the contents of G).
    X : pd.DataFrame
        Features (corresponding to the contents of G).

    Returns
    -------
    G : networkx DiGraph
        Updated graph.
    f_df : dataframe
        Updated basic features.
    X : dataframe
        Updated features.
    """
    id_source = t_df.id_source
    id_destination = t_df.id_destination
    amount = t_df.amount
    nr_alerts = 1 if (t_df.label == 0 or t_df.label == 1) else 0
    nr_reports = 1 if t_df.label == 1 else 0

    source_in_G = id_source in G
    dest_in_G = id_destination in G
    
    the_edge_exists = False
    
    # completely new nodes
    if not source_in_G and not dest_in_G:
        # append basic features for both nodes
        source_features = np.array(
            [[id_source, 0, 1, 0, amount, 0, 1, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0, nr_alerts, nr_reports]])
        destination_features = np.array(
            [[id_destination, 1, 0, amount, 0, 1, 0, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0, nr_alerts, nr_reports]])
        source_features = np.append(source_features, destination_features, axis=0)

        temp_df = pd.DataFrame(source_features, columns=df_columns)
        f_df = pd.concat([f_df, temp_df], ignore_index=True)
        
        # append features for both nodes
        source_features = np.array(
            [[id_source, 0, 1, 0, amount, 0, amount, 0.5, 0, 0, 0, 0, 0, 0, 0]])
        destination_features = np.array(
            [[id_destination, 1, 0, amount, 0, amount, 0, 0.5, 0, 0, 0, 0, 0, 0, 0]])
        source_features = np.append(source_features, destination_features, axis=0)

        temp_df = pd.DataFrame(source_features, columns=df_features)
        X = pd.concat([X, temp_df], ignore_index=True)

    # new source, destination existing
    elif not source_in_G and dest_in_G:
        # basic features
        source_features = np.array(
            [[id_source, 0, 1, 0, amount, 0, 1, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0, nr_alerts, nr_reports]])
        ii_dest = f_df.index[f_df["id"] == id_destination][0]
        f_df["f_degree_in"].iat[ii_dest] += 1
        f_df["f_amount_in"].iat[ii_dest] += amount
        f_df["f_nr_trans_in"].iat[ii_dest] += 1
        f_df["f_ego_nr_nodes"].iat[ii_dest] += 1
        f_df["f_ego_nr_edges"].iat[ii_dest] += 1

        destination_neighbors = []
        destination_neighbors.extend(G.successors(id_destination))
        destination_neighbors.extend(G.predecessors(id_destination))
        n_dest_neighbors = len(list(destination_neighbors))
        if n_dest_neighbors == 1:   # if dest had a single neighbor, now it has two and so it enters its own egonet
            f_df["f_egored_nr_nodes"].iat[ii_dest] += 1

        temp_df = pd.DataFrame(source_features, columns=df_columns)
        f_df = pd.concat([f_df, temp_df], ignore_index=True)

        # features
        source_features = np.array(
            [[id_source, 0, 1, 0, amount, 0, amount, 0.5, 0, 0, 0, 0, 0, 0, 0]])
        temp_df = pd.DataFrame(source_features, columns=df_features)
        X = pd.concat([X, temp_df], ignore_index=True)
        features_from_basic_14(f_df, X, ii_dest)
        
    # new destination, source existing
    elif source_in_G and not dest_in_G:
        destination_features = np.array(
            [[id_destination, 1, 0, amount, 0, 1, 0, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0, nr_alerts, nr_reports]])
        ii_source = f_df.index[f_df["id"] == id_source][0]
        f_df["f_degree_out"].iat[ii_source] += 1
        f_df["f_amount_out"].iat[ii_source] += amount
        f_df["f_nr_trans_out"].iat[ii_source] += 1
        f_df["f_ego_nr_nodes"].iat[ii_source] += 1
        f_df["f_ego_nr_edges"].iat[ii_source] += 1

        source_neighbors = []
        source_neighbors.extend(G.successors(id_source))
        source_neighbors.extend(G.predecessors(id_source))
        n_source_neighbors = len(list(source_neighbors))
        if n_source_neighbors == 1:
            f_df["f_egored_nr_nodes"].iat[ii_source] += 1

        temp_df = pd.DataFrame(destination_features, columns=df_columns)
        f_df = pd.concat([f_df, temp_df], ignore_index=True)

        # features
        destination_features = np.array(
            [[id_destination, 1, 0, amount, 0, amount, 0, 0.5, 0, 0, 0, 0, 0, 0, 0]])
        temp_df = pd.DataFrame(destination_features, columns=df_features)
        X = pd.concat([X, temp_df], ignore_index=True)
        features_from_basic_14(f_df, X, ii_source)

    # both nodes exist
    else:
        source_neighbors = []
        destination_neighbors = []
        source_neighbors.extend(G.successors(id_source))
        source_neighbors.extend(G.predecessors(id_source))
        destination_neighbors.extend(G.successors(id_destination))
        destination_neighbors.extend(G.predecessors(id_destination))

        common_neighbors = list(set(source_neighbors) & set(destination_neighbors))

        ii_source = f_df.index[f_df["id"] == id_source][0]
        ii_dest = f_df.index[f_df["id"] == id_destination][0]
        
        # amount and number of transactions always grow
        f_df["f_amount_out"].iat[ii_source] += amount
        f_df["f_nr_trans_out"].iat[ii_source] += 1
        f_df["f_amount_in"].iat[ii_dest] += amount
        f_df["f_nr_trans_in"].iat[ii_dest] += 1
        
        # if the destination is in egored(source), which also implies the viceversa
        inverse_edge_exists = G.has_edge(id_destination, id_source)
        common_neighbors_exist = (len(common_neighbors) > 0)
        if common_neighbors_exist or inverse_edge_exists:
            f_df["f_egored_amount_out"].iat[ii_source] += amount
            f_df["f_egored_nr_trans_out"].iat[ii_source] += 1
            f_df["f_egored_amount_in"].iat[ii_dest] += amount
            f_df["f_egored_nr_trans_in"].iat[ii_dest] += 1

        # if the edge is new
        the_edge_exists = G.has_edge(id_source, id_destination)
        if not the_edge_exists:
            # degrees always increase
            f_df["f_degree_out"].iat[ii_source] += 1
            f_df["f_degree_in"].iat[ii_dest] += 1
            
            n_source_neighbors = len(list(source_neighbors))
            n_dest_neighbors = len(list(destination_neighbors))
            # if source had a single neighbor, now it has two and so it enters its own egonet
            if n_source_neighbors == 1:
                f_df["f_egored_nr_nodes"].iat[ii_source] += 1
            if n_dest_neighbors == 1:   # same for destination
                f_df["f_egored_nr_nodes"].iat[ii_dest] += 1

            if inverse_edge_exists:
                # the edge joins the egored of the partner
                f_df["f_egored_degree_out"].iat[ii_source] += 1
                f_df["f_egored_degree_in"].iat[ii_dest] += 1

                if not common_neighbors_exist:
                    # the node and the edge are counted in the egored
                    f_df["f_egored_nr_nodes"].iat[ii_source] += 1
                    f_df["f_egored_nr_nodes"].iat[ii_dest] += 1
                    f_df["f_egored_nr_edges"].iat[ii_source] += 1
                    f_df["f_egored_nr_edges"].iat[ii_dest] += 1

                    # the inverse edge must be accounted in the egored
                    f_df["f_egored_degree_in"].iat[ii_source] += 1
                    f_df["f_egored_degree_out"].iat[ii_dest] += 1
                    a = G[id_destination][id_source]["cum_amount"]
                    f_df["f_egored_amount_out"].iat[ii_dest] += a
                    f_df["f_egored_amount_in"].iat[ii_source] += a
                    t = G[id_destination][id_source]["nr_transactions"]
                    f_df["f_egored_nr_trans_out"].iat[ii_dest] += t
                    f_df["f_egored_nr_trans_in"].iat[ii_source] += t
            
            else:
                # if no inverse edge, the node and the edge join the egonet of the partner
                f_df["f_ego_nr_nodes"].iat[ii_source] += 1
                f_df["f_ego_nr_edges"].iat[ii_source] += 1
                f_df["f_ego_nr_nodes"].iat[ii_dest] += 1
                f_df["f_ego_nr_edges"].iat[ii_dest] += 1
            
                if common_neighbors_exist:
                    # the node and the edge join the egored of the partner
                    f_df["f_egored_degree_out"].iat[ii_source] += 1
                    f_df["f_egored_nr_nodes"].iat[ii_source] += 1
                    f_df["f_egored_nr_edges"].iat[ii_source] += 1
                    f_df["f_egored_degree_in"].iat[ii_dest] += 1
                    f_df["f_egored_nr_nodes"].iat[ii_dest] += 1
                    f_df["f_egored_nr_edges"].iat[ii_dest] += 1

            # the common neighbors may enter the egoreds
            for n in common_neighbors:
                ii = f_df.index[f_df["id"] == n][0]
                # egonet and egored number of edges updates
                if not inverse_edge_exists:
                    # the new edge enters the egonet of n
                    f_df["f_ego_nr_edges"].iat[ii] += 1
                    # the edge (n,id_destination) enters the egonet of id_source
                    f_df["f_ego_nr_edges"].iat[ii_source] += 1
                    # the edge (n,id_source) enters the egonet of id_destination
                    f_df["f_ego_nr_edges"].iat[ii_dest] += 1
                
                    # the new edge enters the egored of n
                    f_df["f_egored_nr_edges"].iat[ii] += 1
                    # the edge (n,id_destination) enters the egored of id_source
                    f_df["f_egored_nr_edges"].iat[ii_source] += 1
                    # the edge (n,id_source) enters the egored of id_destination
                    f_df["f_egored_nr_edges"].iat[ii_dest] += 1

                # other egored updates
                if not exist_common_neighbors(n, source_neighbors, G):
                    edge_n_source_exists = G.has_edge(n, id_source)
                    edge_source_n_exists = G.has_edge(id_source, n)

                    if not (edge_n_source_exists and edge_source_n_exists):
                        # if not both edges exist, id_source enters egored(n)
                        f_df["f_egored_nr_nodes"].iat[ii] += 1
                        f_df["f_egored_nr_edges"].iat[ii] += 1
                        # and n enters egored(id_source)
                        f_df["f_egored_nr_nodes"].iat[ii_source] += 1
                        f_df["f_egored_nr_edges"].iat[ii_source] += 1
                    
                    # the edge (n,id_source) contributes to egoreds of both nodes (and wasn't before)
                    if edge_n_source_exists and not edge_source_n_exists:
                        f_df["f_egored_degree_out"].iat[ii] += 1
                        f_df["f_egored_degree_in"].iat[ii_source] += 1
                        a = G[n][id_source]["cum_amount"]
                        f_df["f_egored_amount_out"].iat[ii] += a
                        f_df["f_egored_amount_in"].iat[ii_source] += a
                        t = G[n][id_source]["nr_transactions"]
                        f_df["f_egored_nr_trans_out"].iat[ii] += t
                        f_df["f_egored_nr_trans_in"].iat[ii_source] += t
                        
                    # same for (id_source,n)
                    if edge_source_n_exists and not edge_n_source_exists:
                        f_df["f_egored_degree_in"].iat[ii] += 1
                        f_df["f_egored_degree_out"].iat[ii_source] += 1
                        a = G[id_source][n]["cum_amount"]
                        f_df["f_egored_amount_in"].iat[ii] += a
                        f_df["f_egored_amount_out"].iat[ii_source] += a
                        t = G[id_source][n]["nr_transactions"]
                        f_df["f_egored_nr_trans_in"].iat[ii] += t
                        f_df["f_egored_nr_trans_out"].iat[ii_source] += t

                # same as above for id_destination        
                if not exist_common_neighbors(n, destination_neighbors, G):
                    edge_n_dest_exists = G.has_edge(n, id_destination)
                    edge_dest_n_exists = G.has_edge(id_destination, n)

                    if not (edge_n_dest_exists and edge_dest_n_exists):
                        # id_destination enters egored(n)
                        f_df["f_egored_nr_nodes"].iat[ii] += 1
                        f_df["f_egored_nr_edges"].iat[ii] += 1
                        # n enters egored(id_destination)
                        f_df["f_egored_nr_nodes"].iat[ii_dest] += 1
                        f_df["f_egored_nr_edges"].iat[ii_dest] += 1
                    
                    # the edge (n,id_destination) contributes to egoreds of both nodes (and wasn't before)
                    if edge_n_dest_exists and not edge_dest_n_exists:
                        f_df["f_egored_degree_out"].iat[ii] += 1
                        f_df["f_egored_degree_in"].iat[ii_dest] += 1
                        a = G[n][id_destination]["cum_amount"]
                        f_df["f_egored_amount_out"].iat[ii] += a
                        f_df["f_egored_amount_in"].iat[ii_dest] += a
                        t = G[n][id_destination]["nr_transactions"]
                        f_df["f_egored_nr_trans_out"].iat[ii] += t
                        f_df["f_egored_nr_trans_in"].iat[ii_dest] += t
                        
                    # same for (id_destination,n)
                    if edge_dest_n_exists and not edge_n_dest_exists:
                        f_df["f_egored_degree_in"].iat[ii] += 1
                        f_df["f_egored_degree_out"].iat[ii_dest] += 1
                        a = G[id_destination][n]["cum_amount"]
                        f_df["f_egored_amount_in"].iat[ii] += a
                        f_df["f_egored_amount_out"].iat[ii_dest] += a
                        t = G[id_destination][n]["nr_transactions"]
                        f_df["f_egored_nr_trans_in"].iat[ii] += t
                        f_df["f_egored_nr_trans_out"].iat[ii_dest] += t
                
                # update features, now that node n info has been modified
                features_from_basic_14(f_df, X, ii)
                
        # update source and destination features
        features_from_basic_14(f_df, X, ii_source)
        features_from_basic_14(f_df, X, ii_dest)
            
    # add edge info to the graph
    if the_edge_exists:
        # alter the edge attributes if there is an edge between the nodes
        G[id_source][id_destination]["nr_transactions"] += 1
        G[id_source][id_destination]["cum_amount"] += amount
        G[id_source][id_destination]["nr_alerts"] += nr_alerts
        G[id_source][id_destination]["nr_reports"] += nr_reports
    else:
        # add edge with attributes if there is not one
        G.add_edge(id_source, id_destination, cum_amount=amount,
                   nr_transactions=1, nr_alerts=nr_alerts, nr_reports=nr_reports)

    return G, f_df, X
